@@ -71,12 +71,14 @@ static void *zalloc(size_t size)
 
 static const char **ips;
 static int ips_idx;
+static int func_depth;
 static int current_pid = -1;
 
 struct stack_save {
 	struct stack_save	*next;
 	const char		**ips;
 	int			ips_idx;
+	int			func_depth;
 	int			pid;
 };
 
@@ -86,6 +88,7 @@ static void reset_stack(void)
 {
 	current_pid = -1;
 	ips_idx = 0;
+	func_depth = 0;
 	/* Don't free here, it may be saved */
 	ips = NULL;
 }
@@ -100,6 +103,7 @@ static void save_stack(void)
 
 	stack->pid = current_pid;
 	stack->ips_idx = ips_idx;
+	stack->func_depth = func_depth;
 	stack->ips = ips;
 
 	stack->next = saved_stacks;
@@ -127,6 +131,7 @@ static void restore_stack(int pid)
 
 	current_pid = stack->pid;
 	ips_idx = stack->ips_idx;
+	func_depth = stack->func_depth;
 	free(ips);
 	ips = stack->ips;
 	free(stack);
@@ -318,8 +323,84 @@ process_function(struct pevent *pevent, struct pevent_record *record)
 }
 
 static void
-process_function_graph(struct pevent *pevent, struct pevent_record *record)
+process_function_graph_entry(struct pevent *pevent, struct pevent_record *record)
 {
+	unsigned long long depth;
+	unsigned long long ip;
+	unsigned long long val;
+	const char *func;
+	int pid;
+	int ret;
+
+	ret = pevent_read_number_field(common_pid_field, record->data, &val);
+	if (ret < 0)
+		die("no pid field for function graph entry?");
+
+	ret = pevent_read_number_field(function_graph_entry_func_field,
+				       record->data, &ip);
+	if (ret < 0)
+		die("no ip field for function graph entry?");
+
+	ret = pevent_read_number_field(function_graph_entry_depth_field,
+				       record->data, &depth);
+	if (ret < 0)
+		die("no parent ip field for function entry?");
+
+	pid = val;
+
+	func = pevent_find_function(pevent, ip);
+
+	if (current_pid >= 0 && pid != current_pid) {
+		save_stack();
+		restore_stack(pid);
+	}
+
+	current_pid = pid;
+
+	if (depth != ips_idx) {
+		save_call_chain(pid, ips, ips_idx, 0);
+		while (ips_idx > depth)
+			pop_stack_func();
+	}
+
+	func_depth = depth;
+
+	push_stack_func(func);
+}
+
+static void
+process_function_graph_exit(struct pevent *pevent, struct pevent_record *record)
+{
+	unsigned long long depth;
+	unsigned long long val;
+	int pid;
+	int ret;
+
+	ret = pevent_read_number_field(common_pid_field, record->data, &val);
+	if (ret < 0)
+		die("no pid field for function graph exit?");
+
+	ret = pevent_read_number_field(function_graph_exit_depth_field,
+				       record->data, &depth);
+	if (ret < 0)
+		die("no parent ip field for function?");
+
+	pid = val;
+
+	if (current_pid >= 0 && pid != current_pid) {
+		save_stack();
+		restore_stack(pid);
+	}
+
+	current_pid = pid;
+
+	if (ips_idx != depth) {
+		save_call_chain(pid, ips, ips_idx, 0);
+		while (ips_idx > depth)
+			pop_stack_func();
+	}
+
+	func_depth = depth - 1;
 }
 
 static int pending_pid = -1;
@@ -503,9 +584,11 @@ process_record(struct pevent *pevent, struct pevent_record *record)
 	if (type == function_type)
 		return process_function(pevent, record);
 
-	if (type == function_graph_entry_type ||
-	    type == function_graph_exit_type)
-		return process_function_graph(pevent, record);
+	if (type == function_graph_entry_type)
+		return process_function_graph_entry(pevent, record);
+
+	if (type == function_graph_exit_type)
+		return process_function_graph_exit(pevent, record);
 
 	if (type == kernel_stack_type)
 		return process_kernel_stack(pevent, record);
